
/*
 *  finddup is a simple utility find duplicated files, files common to
 *  several directories, or files present in one directory and not in
 *  another one.
 *
 *  Copyright (c) 2010 Francois Fleuret
 *  Written by Francois Fleuret <francois@fleuret.org>
 *
 *  This file is part of finddup.
 *
 *  finddup is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  finddup is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 *  License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with finddup.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define VERSION_NUMBER "0.6"

#define _BSD_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <locale.h>
#include <getopt.h>
#include <fcntl.h>

/* 1M really helps compared to 64k */
#define READ_BUFFER_SIZE (1024 * 1024)

typedef int64_t size_sum_t;

/* Yeah, global variables! */

int ignore_dotfiles = 0; /* 1 means ignore files and directories
                            starting with a dot */

int ignore_empty_files = 0; /* 1 means ignore empty files */

int show_realpaths = 0; /* 1 means ignore files and directories
                            starting with a dot */

int show_progress = 0; /* 1 means show a progress bar when we are in a
                          tty */

int show_hits = 1; /* 1 means to show the files from dir2
                      corresponding to the ones from dir1 */

int show_groups = 1; /* 1 means to show the group IDs when printing
                        file names */

int ignore_same_inode = 0; /* 1 means that comparison between two file
                              with same inode will always be false */

int tty_width = -1; /* Positive value means what width to use to show
                       the progress bar */

/********************************************************************/

/* malloc with error checking.  */

void *safe_malloc(size_t n) {
  void *p = malloc(n);
  if (!p && n != 0) {
    printf("Can not allocate memory: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  return p;
}

/********************************************************************/

int ignore_entry(const char *name) {
  return
    strcmp(name, ".") == 0 ||
    strcmp(name, "..") == 0 ||
    (ignore_dotfiles && name[0] == '.');
}

void print_size_sum(size_sum_t s) {
  char tmp[100];
  char *a = tmp + sizeof(tmp)/sizeof(char);
  *(--a) = '\0';
  if(s) {
    while(s) {
      *(--a) = s%10 + '0';
      s /= 10;
    }
  } else {
    *(--a) = '0';
  }
  printf(a);
}

/**********************************************************************/

struct file_with_size {
  char *filename;
  size_t size;
  ino_t inode;
  struct file_with_size *next;
  int group_id, dir_id;
};

void file_list_delete(struct file_with_size *head) {
  struct file_with_size *next;
  while(head) {
    next = head->next;
    free(head->filename);
    free(head);
    head = next;
  }
}

int file_list_length(struct file_with_size *head) {
  int l = 0;
  while(head) {
    l++;
    head = head->next;
  }
  return l;
}

/**********************************************************************/

int same_content(struct file_with_size *f1, struct file_with_size *f2,
                 char *buffer1, char *buffer2) {
  int fd1, fd2, s1, s2;

  fd1 = open(f1->filename, O_RDONLY);
  fd2 = open(f2->filename, O_RDONLY);

  if(fd1 >= 0 && fd2 >= 0) {
    while(1) {
      s1 = read(fd1, buffer1, READ_BUFFER_SIZE);
      s2 = read(fd2, buffer2, READ_BUFFER_SIZE);

      if(s1 < 0 || s2 < 0) {
        close(fd1);
        close(fd2);
        return 0;
      }

      if(s1 == s2) {
        if(s1 == 0) {
          close(fd1);
          close(fd2);
          return 1;
        } else {
          if(memcmp(buffer1, buffer2, s1)) {
            close(fd1);
            close(fd2);
            return 0;
          }
        }
      } else {
        fprintf(stderr,
                "Different read size without error on files of same size.\n");
        exit(EXIT_FAILURE);
      }
    }
  } else {

    if(fd1 < 0) {
      fprintf(stderr,
              "Can not open \"%s\" error: %s\n",
              f1->filename,
              strerror(errno));
    }

    if(fd2 < 0) {
      fprintf(stderr,
              "Can not open \"%s\" error: %s\n",
              f2->filename,
              strerror(errno));
    }
    exit(EXIT_FAILURE);
  }
}

int same_files(struct file_with_size *f1, struct file_with_size *f2,
               char *buffer1, char *buffer2) {
  if(ignore_same_inode && f1->inode == f2->inode) {
    return 0;
  }
  return f1->size == f2->size && same_content(f1, f2, buffer1, buffer2);
}

/**********************************************************************/

struct file_with_size *scan_directory(struct file_with_size *tail,
                                      const char *name) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat sb;
  struct file_with_size *tmp;
  char subname[PATH_MAX + 1];

  if(lstat(name, &sb) != 0) {
    fprintf(stderr, "Can not stat \"%s\": %s\n", name, strerror(errno));
    exit(EXIT_FAILURE);
  }

  if(S_ISLNK(sb.st_mode)) {
    return tail;
  }

  dir = opendir(name);

  if(dir) {
    while((dir_e = readdir(dir))) {
      if(!ignore_entry(dir_e->d_name)) {
        snprintf(subname, PATH_MAX, "%s/%s", name, dir_e->d_name);
        tail = scan_directory(tail, subname);
      }
    }
    closedir(dir);
  } else {
    if(S_ISREG(sb.st_mode)) {
      if(!ignore_entry(name)) {
        if(!ignore_empty_files || sb.st_size > 0) {
          tmp = safe_malloc(sizeof(struct file_with_size));
          tmp->next = tail;
          tmp->filename = strdup(name);
          tmp->size = sb.st_size;
          tmp->inode = sb.st_ino;
          tmp->group_id = -1;
          tmp->dir_id = -1;
          tail = tmp;
        }
      }
    }
  }

  return tail;
}

void print_file(struct file_with_size *node) {
  char tmp[PATH_MAX + 1];
  if(show_realpaths) {
    if(show_groups) {
      realpath(node->filename, tmp);
      printf("%d %s\n", node->group_id, tmp);
    } else {
      realpath(node->filename, tmp);
      printf("%s\n", tmp);
    }
  } else {
    if(show_groups) {
      printf("%d %s\n", node->group_id, node->filename);
    } else {
      printf("%s\n", node->filename);
    }
  }
}

int compare_nodes(const void *x1, const void *x2) {
  const struct file_with_size **f1, **f2;

  f1 = (const struct file_with_size **) x1;
  f2 = (const struct file_with_size **) x2;

  if((*f1)->group_id < (*f2)->group_id) {
    return -1;
  } else if((*f1)->group_id > (*f2)->group_id) {
    return 1;
  } else {
    if((*f1)->dir_id < (*f2)->dir_id) {
      return -1;
    } else if((*f1)->dir_id > (*f2)->dir_id) {
      return 1;
    } else {
      return 0;
    }
  }
}


void print_result(struct file_with_size *list1, struct file_with_size *list2) {
  struct file_with_size *node1, *node2;
  struct file_with_size **nodes;
  int nb, n;

  nb = 0;
  for(node1 = list1; node1; node1 = node1->next) {
    if(node1->group_id >= 0) { nb++; }
  }

  if(list2) {
    for(node2 = list2; node2; node2 = node2->next) {
      if(node2->group_id >= 0) { nb++; }
    }
  }

  nodes = safe_malloc(nb * sizeof(struct file_with_size *));

  n = 0;
  for(node1 = list1; node1; node1 = node1->next) {
    if(node1->group_id >= 0) {
      nodes[n++] = node1;
    }
  }

  if(list2) {
    for(node2 = list2; node2; node2 = node2->next) {
      if(node2->group_id >= 0) {
        nodes[n++] = node2;
      }
    }
  }

  qsort(nodes, nb, sizeof(struct file_with_size *), compare_nodes);

  for(n = 0; n < nb; n++) {
    if(!show_groups && n > 0 && nodes[n]->group_id != nodes[n-1]->group_id) {
      printf("\n");
    }
    print_file(nodes[n]);
  }

  free(nodes);
}

void print_progress(int max, int n, int *pp) {
  int p, k;
  int width;
  if(show_progress && tty_width > 0) {
    width = tty_width - 7;
    p = (width * n) / max;
    if(p > *pp) {
      for(k = 0; k < p; k++) {
        fprintf(stderr, "+");
      }
      for(; k < width; k++) {
        fprintf(stderr, "-");
      }
      *pp = p;
      p = (100 * n) / max;
      fprintf(stderr, " [% 3d%%]\r", p);
    }
  }
}

void start(const char *dirname1, const char *dirname2) {
  struct file_with_size *list1, *list2;
  struct file_with_size *node1, *node2;
  int not_in, found;
  int k, pp, l1, n;

  char *buffer1 = safe_malloc(sizeof(char) * READ_BUFFER_SIZE);
  char *buffer2 = safe_malloc(sizeof(char) * READ_BUFFER_SIZE);

  not_in = 0;

  if(show_progress) {
    fprintf(stderr, "Scanning %s ... ", dirname1);
  }

  list1 = scan_directory(0, dirname1);

  if(dirname2) {
    if(strncmp(dirname2, "not:", 4) == 0) {
      not_in = 1;
      /* We should show groups even in that mode. However they are not
         properly calculated for now, so we force it off. */
      show_groups = 0;
      dirname2 += 4;
    }
    if(show_progress) {
      fprintf(stderr, "%s ... ", dirname2);
    }
    list2 = scan_directory(0, dirname2);
  } else {
    list2 = list1;
  }

  if(show_progress) {
    fprintf(stderr, "done.\n");
  }

  k = 0;
  pp = -1;
  n = 0;
  l1 = file_list_length(list1);

  if(not_in) {
    for(node1 = list1; node1; node1 = node1->next) {
      print_progress(l1, n, &pp);
      n++;

      found = 0;

      for(node2 = list2; !found && node2; node2 = node2->next) {
        if(same_files(node1, node2, buffer1, buffer2)) {
          found = 1;
        }
      }

      if(!found) {
        if(show_realpaths) {
          printf("%s\n", realpath(node1->filename, 0));
        } else {
          printf("%s\n", node1->filename);
        }
      }
    }

  } else {
    for(node1 = list1; node1; node1 = node1->next) {
      print_progress(l1, n, &pp);
      n++;

      for(node2 = list2; node2; node2 = node2->next) {
        if(node1->group_id < 0 || node2->group_id < 0) {
          if(same_files(node1, node2, buffer1, buffer2)) {
            if(node1->group_id < 0) {
              if(node2->group_id >= 0) {
                node1->group_id = node2->group_id;
              } else {
                node1->group_id = k;
                node1->dir_id = 1;
                k++;
              }
            }
            if(node2->group_id < 0) {
              node2->group_id = node1->group_id;
              node2->dir_id = 2;
            }
          }
        }
      }
    }
  }

  if(show_progress) {
    fprintf(stderr, "\n");
  }

  if(dirname2) {
    print_result(list1, list2);
    file_list_delete(list1);
    file_list_delete(list2);
  } else {
    print_result(list1, 0);
    file_list_delete(list1);
  }

  free(buffer1);
  free(buffer2);
}

void print_help(FILE *out) {
  fprintf(out, "Usage: finddup [OPTION]... DIR1 [[not:]DIR2]\n");
  fprintf(out, "Version %s (%s)\n", VERSION_NUMBER, UNAME);
  fprintf(out, "Without DIR2, lists duplicated files found in DIR1. With DIR2, lists files common to both directories. With the not: prefix, lists files found in DIR1 which do not exist in DIR2.\n");
  fprintf(out, "\n");
  fprintf(out, "   -h   show this help\n");
  fprintf(out, "   -d   ignore dot files and directories\n");
  fprintf(out, "   -0   ignore empty files\n");
  fprintf(out, "   -c   do not show which files in DIR2 corresponds to those in DIR1\n");
  fprintf(out, "   -g   do not show the file groups\n");
  fprintf(out, "   -p   show progress\n");
  fprintf(out, "   -r   show the real file paths\n");
  fprintf(out, "   -i   consider files with same inode as different\n");
  fprintf(out, "\n");
  fprintf(out, "Report bugs and comments to <francois@fleuret.org>\n");
}

/**********************************************************************/

int main(int argc, char **argv) {
  int c;
  struct file_with_size *root;
  struct winsize win;

  root = 0;

  setlocale (LC_ALL, "");

  while (1) {
    c = getopt(argc, argv, "hircgd0p");
    if (c == -1)
      break;

    switch (c) {

    case 'h':
      print_help(stdout);
      exit(EXIT_SUCCESS);

      break;

    case 'd':
      ignore_dotfiles = 1;
      break;

    case '0':
      ignore_empty_files = 1;
      break;

    case 'r':
      show_realpaths = 1;
      break;

    case 'i':
      ignore_same_inode = 1;
      break;

    case 'g':
      show_groups = 0;
      break;

    case 'p':
      show_progress = 1;
      break;

    case 'c':
      show_hits = 0;
      break;

    default:
      exit(EXIT_FAILURE);
    }
  }

  if(show_progress &&
     isatty(STDOUT_FILENO) &&
     !ioctl (STDOUT_FILENO, TIOCGWINSZ, (char *) &win)) {
    tty_width = win.ws_col;
  }

  if(optind + 2 == argc) {
    start(argv[optind], argv[optind + 1]);
  } else if(optind + 1 == argc) {
    ignore_same_inode = 1;
    start(argv[optind], 0);
  } else {
    print_help(stderr);
    exit(EXIT_FAILURE);
  }

  exit(EXIT_SUCCESS);
}
